from datetime import datetime
import pandas as pd
from pandas import to_datetime
import fitbit_data_pipeline.Utility as util

log = util.get_logger()

class Participant:
    def __init__(self, participant_id: str, age: int, study_period: tuple, collection_days: list):
        self.participant_id = participant_id
        self.assigned_fitbit = None
        self.age = age
        self.study_period = study_period  # (start_date, end_date)
        self.collection_days = collection_days
        self.is_consecutive_days = self.check_consecutive()
        self.token = None  # Fitbit authorisation token
        self.is_authorized = False  # Authorization status

    def check_consecutive(self):
        study_start = to_datetime(self.study_period[0])
        study_end = to_datetime(self.study_period[1])
        collection_dates = sorted(pd.to_datetime(self.collection_days))
        self.collection_days = [date.strftime('%Y-%m-%d') for date in collection_dates]
        within_range = all(study_start <= date <= study_end for date in collection_dates)
        if not within_range:
            raise ValueError(f"One or more of Collection dates out of range for  {self.participant_id}. Please check!")
            #print(f"One or more of Collection dates out of range for  {self.participant_id}. Please check!")
            #return False
        for i in range(1, len(collection_dates)):
            if (collection_dates[i] - collection_dates[i - 1]).days != 1:
                return False #not consecutive days
        return True #consecutive days


class Device:
    def __init__(self, fitbit_user_id: str, model: str):
        self.device_id = fitbit_user_id
        self.model = model
        self.current_session = None  # The current session (assigned participant) for the device

    def assign_to_participant(self, session):
        self.current_session = session

    def release_device(self):
        self.current_session = None


class Session:
    def __init__(self, participant: Participant, device: Device, start_time: datetime, end_time: datetime = None):
        self.participant = participant
        self.device = device
        self.start_time = start_time
        self.end_time = end_time

    def end_session(self, end_time: datetime):
        self.end_time = end_time


import json
from datetime import datetime
from pathlib import Path

from fitbit_data_pipeline.Classes.Participant import Device
from fitbit_data_pipeline.Classes.Participant import Participant
from fitbit_data_pipeline.Classes.Participant import Session
from dotenv import set_key
import os
import fitbit_data_pipeline.Utility as util


env_file = ('');os.path.join(os.path.dirname(os.path.dirname(__file__)), '.')
log = util.get_logger()

class ParticipantManager:
    def __init__(self, token_dir="tokens"):
        self.participants = {}
        self.devices = {}
        self.sessions = []
        self.tokens_dir = Path(token_dir)
        self.tokens_dir.mkdir(exist_ok=True)
        self.participants_error = []



    def add_participant(self, participant: Participant):
        self.participants[participant.participant_id] = participant
        #self.load_token(participant.participant_id)

    def add_device(self, device: Device):
        self.devices[device.device_id] = device

    def assign_device_to_participant(self, participant_id: str, device_id: str):
        participant = self.participants.get(participant_id)
        device = self.devices.get(device_id)

        if participant is None:
            log.error(f"Participant with ID {participant_id} does not exist.")
            raise ValueError(f"Participant with ID {participant_id} does not exist.")
        if device is None:
            log.error(f"Device with Fitbit ID {device_id} does not exist.")
            raise ValueError(f"Device with Fitbit ID {device_id} does not exist.")
        if device.current_session is not None:
            log.error(f"Device {device_id} is currently in use by another participant.")
            raise ValueError(f"Device {device_id} is currently in use by another participant.")

        # Check if the participant already has an active session
        existing_sessions = self.get_participant_session(participant_id)
        if existing_sessions:
            log.error(f"There is an active session for {participant_id}")
            raise ValueError(f"There is an active session for {participant_id}")

        session_start_time = participant.study_period[0]

        session = Session(participant, device, session_start_time) #new session created for a new participant
        self.sessions.append(session)
        device.assign_to_participant(session)
        #participant.
        log.info(f"User {participant_id} has been assigned to {device_id} for session starting at {session_start_time}")
        #print(f"User {participant_id} has been assigned to {device_id} for session starting at {session_start_time}")

    def end_session(self, session: Session, end_time: datetime):
        session.end_session(end_time)
        session.device.release_device()

    def get_active_sessions(self):
        return [s for s in self.sessions if s.end_time is None]

    def get_participant_session(self, participant_id: str):
        for session in self.sessions:
            if session.participant.participant_id == participant_id:
                return session
        return None

    #@staticmethod
    # def save_token(participant: Participant):
    #     token_data = {
    #         'participant_id': participant.participant_id,
    #         'token': participant.token
    #     }
    #
    #     # Load existing tokens from the file
    #     p_tokens_str = os.getenv("participants_tokens","{}")
    #     token_json = json.loads(p_tokens_str)
    #
    #     # Update token for the current participant
    #     token_json[participant.participant_id] = token_data
    #
    #     updated_tokens = json.dumps(token_json)
    #     set_key('.env',"participants_tokens",updated_tokens)
    #     print(f"Token for participant {participant.participant_id} saved.")
    #
    # @staticmethod
    # def load_token(participant: Participant):
    #     try:
    #         tokens_str = os.getenv("participants_tokens","{}")
    #         print("Token string", tokens_str)
    #         token_dic = json.loads(tokens_str)
    #         print("Token string dic", token_dic)
    #         # Check if token exists for the participant
    #         if participant.participant_id in token_dic:
    #             participant.token = token_dic[participant.participant_id]['token']
    #             print(participant.participant_id, participant.token)
    #             participant.is_authorized = True
    #             print(f"Loaded token for participant {participant.participant_id}")
    #         else:
    #             print(f"No token found for participant {participant.participant_id}, re-authorization needed.")
    #
    #     except FileNotFoundError:
    #         print("No tokens file found. All participants need to be authorized.")

    def locate_token(self, pid):
        return self.tokens_dir / f"{pid}_token.json"

    def save_token(self, participant):
        path = self.locate_token(participant.participant_id)
        with open(path, "w") as file:
            json.dump(participant.token, file)
        log.info(f"Token successfully saved for user {participant.participant_id}")
        #print(f"Token successfully saved for user {participant.participant_id}")

    def load_token(self, participant):
        user_token = self.locate_token(participant.participant_id)
        if not user_token.exists():
            log.info(f"Token for user {participant.participant_id} not found!")
            #print(f"Token for user {participant.participant_id} not found!")
            return None
        else:
            with open(user_token, "r") as file:
                participant.token = json.load(file)
                return True



import os
import time
import webbrowser
from datetime import datetime
import pandas as pd
from requests_oauthlib import OAuth2Session
from config import FITBIT_AUTH_URL
from .Participant import Participant
from .Participant import Session
from fitbit_data_pipeline.Classes.PManager import ParticipantManager
import config as cfg
import fitbit_data_pipeline.Utility as util

log = util.get_logger()


class DataCollector:
    FITBIT_API_BASE_URL = cfg.FITBIT_API_BASE_URL
    FITBIT_SLEEP_ENDPOINT_RANGE = cfg.FITBIT_SLEEP_ENDPOINT_RANGE
    FITBIT_SLEEP_ENDPOINT_EACH = cfg.FITBIT_SLEEP_ENDPOINT_EACH
    FITBIT_STEPS_ENDPOINT_EACH = cfg.FITBIT_STEPS_ENDPOINT_EACH
    FITBIT_HR_ENDPOINT_EACH = cfg.FITBIT_HR_ENDPOINT_EACH
    FITBIT_AUTH_URL = cfg.FITBIT_AUTH_URL
    TOKEN_URL = cfg.FITBIT_TOKEN_URL
    SCOPES = cfg.SCOPES

    def __init__(self, client_id: str, client_secret: str, redirect_uri: str, token: dict = None):
        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_uri = redirect_uri
        self.token = token

    def is_token_expired(self, token: dict) -> bool:
        if 'expires_at' not in token:
            return True

        return time.time() > token['expires_at']

    def ensure_token_valid(self, participant, pm):
        if self.token and 'expires_at' in self.token and datetime.now().timestamp() > self.token['expires_at']:
            log.info(f"Access token expired for {participant.participant_id}, refreshing...")
            #print(f"Access token expired for {participant.participant_id}, refreshing...")
            self.refresh_access_token(participant, pm)

    def refresh_access_token(self, participant, pm):
        extra = {'client_id': self.client_id, 'client_secret': self.client_secret}
        oauth2_session = OAuth2Session(self.client_id, token=self.token)
        self.token = oauth2_session.refresh_token(self.TOKEN_URL, **extra)
        participant.token = self.token
        pm.save_token(participant)

        log.info(f"Access token refreshed for {participant.participant_id}:", self.token)
        print("Access token refreshed:", self.token)

    def get_fitbit_oauth(self, participant: Participant, participant_manager: ParticipantManager):
        # print("part token here", participant.token)
        participant_manager.load_token(participant)
        #print(f"Token expired: {self.is_token_expired(participant.token)}")
        if participant.token and not self.is_token_expired(participant.token):
            log.info(f"Participant {participant.participant_id} already authorized with a valid token.")
            print(f"Participant {participant.participant_id} already authorized with a valid token.")
            participant.is_authorized = True
            return  # No need to re-authorize

        fitbit = OAuth2Session(self.client_id, redirect_uri=self.redirect_uri, scope=self.SCOPES)
        authorization_url, state = fitbit.authorization_url(FITBIT_AUTH_URL)
        #print(f"Please go here and authorize participant {participant.participant_id}: {authorization_url}")
        print(f"Launching the browser to authorise user {participant.participant_id}")

        webbrowser.open("https://www.fitbit.com/logout")
        time.sleep(5)
        print(f"Press log into the fitbit account for user {participant.participant_id}. Then press the enter key")
        input()
        webbrowser.open(authorization_url)
        timeout = 30
        start_time = time.time()
        redirect_response = None
        auth_path = os.path.abspath("auth_code.txt")
        print("Waiting for the redirect URL file to be written")
        while time.time() - start_time < timeout:
            if os.path.exists(auth_path):
                time.sleep(2)
                with open(auth_path, "r") as file:
                    response = file.read()
                    if "code=" in response and "state=" in response:
                        redirect_response = response
                        #print("redirect URL now written!")
                        print(f"redirect URL {redirect_response}")
                        # Fetch token and store it in the participant object
                        token = fitbit.fetch_token(self.TOKEN_URL, authorization_response=redirect_response,
                                                   client_secret=self.client_secret)
                        participant.token = token
                        print(f"Assigned fitbit id: {participant.assigned_fitbit}")
                        print(f"Returned token user id: {token['user_id']}")
                        if participant.assigned_fitbit == token['user_id']:
                            participant.is_authorized = True
                            participant_manager.save_token(participant)
                            log.info(f"Authorization successful for participant {participant.participant_id}")
                            print(f"Authorization successful for participant {participant.participant_id}")
                        else:
                            participant_manager.participants_error.append(participant.participant_id)
                            log.warning(f"Fitbit profile and participant device mis-match for user {participant.participant_id}/nThis user will be skipped")
                            print(f"Fitbit profile and participant device mis-match for user {participant.participant_id}/nThis user will be skipped")
                        break
            time.sleep(2)
        if not redirect_response:
            raise TimeoutError("Timeout: Did not receive auth code within expected time.")


    def collect_intraday_data(self, session: Session, pm: ParticipantManager):
        user_id = session.device.device_id
        #print('user here', user_id)
        if not session.participant.is_authorized:
            log.info(f"User {user_id} is not authorized. Need to authorise user")
            print(f"User {user_id} is not authorized. Need to authorise user")
            self.get_fitbit_oauth(session.participant)

        self.ensure_token_valid(session.participant, pm)  # what if token became invalid at this second????
        fitbit = OAuth2Session(self.client_id, token=session.participant.token)
        steps_df = []
        heart_rate_df = []
        step_counts = None
        heartrate = None
        if "activity" in self.SCOPES:
            for sleep_date in session.participant.collection_days:
                steps_url = self.FITBIT_API_BASE_URL + self.FITBIT_STEPS_ENDPOINT_EACH.format(fitbit_device_id=user_id,
                                                                                              sleep_date=sleep_date)
                response = fitbit.get(steps_url)
                if response.status_code == 403:
                    log.error(f"insufficient permissions for intraday data for {session.participant.participant_id} - activity")
                    raise Exception(f"insufficient permissions for intraday data - activity")
                if response.status_code != 200:
                    log.error(f"Error fetching sleep data for participant {session.participant.participant_id} and date {sleep_date}: {response.status_code} - {response.text}")
                    raise Exception(
                        f"Error fetching sleep data for participant {session.participant.participant_id} and date {sleep_date}: {response.status_code} - {response.text}")
                steps_data = response.json()
                date = steps_data["activities-steps"][0]["dateTime"]
                intraday_data = steps_data["activities-steps-intraday"]["dataset"]
                time_steps = [(entry["time"], entry["value"]) for entry in intraday_data]

                df = pd.DataFrame(time_steps, columns=["time", "activity_steps"])
                df["date"] = date
                df['pid'] = session.participant.participant_id
                df = df[['pid', 'date', 'time', "activity_steps"]]
                steps_df.append(df)
            step_counts = pd.concat(steps_df, ignore_index=True) if len(steps_df) > 0 else None

                ######heart rate ##################
            if "heartrate" in self.SCOPES:
                for sleep_date in session.participant.collection_days:
                    hr_url = self.FITBIT_API_BASE_URL + self.FITBIT_HR_ENDPOINT_EACH.format(fitbit_device_id=user_id,
                                                                                            sleep_date=sleep_date)
                    response = fitbit.get(hr_url)
                    if response.status_code == 403:
                        raise Exception(f"insufficient permissions for intraday data - heartrate")

                    if response.status_code != 200:
                        raise Exception(
                            f"Error fetching sleep data for participant {session.participant.participant_id} and date {sleep_date}: {response.status_code} - {response.text}")
                    hr_data = response.json()
                    date = hr_data["activities-heart"][0]["dateTime"]
                    # Extract time and steps from activities-steps-intraday
                    intraday_data = hr_data["activities-heart-intraday"]["dataset"]
                    time_hr = [(entry["time"], entry["value"]) for entry in intraday_data]

                    df = pd.DataFrame(time_hr, columns=["time", "heart_rate"])
                    df["date"] = date
                    df['pid'] = session.participant.participant_id

                    df = df[['pid', 'date', 'time', "heart_rate"]]
                    heart_rate_df.append(df)
                heartrate = pd.concat(heart_rate_df, ignore_index=True) if len(heart_rate_df) > 0 else None
        return step_counts, heartrate

    def collect_sleep_data(self, session: Session, pm: ParticipantManager):
        user_id = session.device.device_id
        print('user here', user_id)
        if not session.participant.is_authorized:
            log.info(f"User {user_id} is not authorized. Initiating authorization flow...")
            print(f"User {user_id} is not authorized. Initiating authorization flow...")
            self.get_fitbit_oauth(session.participant)

        self.ensure_token_valid(session.participant, pm)
        fitbit = OAuth2Session(self.client_id, token=session.participant.token)
        sleep_data = ''
        if session.participant.is_consecutive_days:
            start_date = session.participant.collection_days[0]
            end_date = session.participant.collection_days[-1]
            #print(f'Data collection days: {start_date} to {end_date}')
            url = self.FITBIT_API_BASE_URL + self.FITBIT_SLEEP_ENDPOINT_RANGE.format(fitbit_device_id=user_id,
                                                                                     start_date=start_date,
                                                                                     end_date=end_date)
            #print('URL', url)
            response = fitbit.get(url)
            #print("response", response.text)
            #print(response.json())
            if response.status_code != 200:
                log.error(f"Error fetching sleep data: {response.status_code} - {response.text}")
                raise Exception(f"Error fetching sleep data: {response.status_code} - {response.text}")
            sleep_data = response.json()['sleep']

            util.dump_data(session.participant.participant_id, start_date, end_date, sleep_data)
        else:
            all_sleep_data = []

            for sleep_date in session.participant.collection_days:
                url = self.FITBIT_API_BASE_URL + self.FITBIT_SLEEP_ENDPOINT_EACH.format(fitbit_device_id=user_id,
                                                                                        sleep_date=sleep_date)
                response = fitbit.get(url)
                #print('Sleep Date ' + sleep_date)
                #print(response.json())
                #print(response.status_code)
                if response.status_code != 200:
                    raise Exception(
                        f"Error fetching sleep data for participant {session.participant.participant_id} and date {sleep_date}: {response.status_code} - {response.text}")
                if len(response.json()['sleep']) > 0:
                    one_sleep_data = response.json()['sleep'][0]
                    all_sleep_data.append(one_sleep_data)
                    sleep_data = all_sleep_data
                else:
                    log.info(f"no sleep data for user {user_id} on {sleep_date}")
                    print(f"no sleep data for user {user_id} on {sleep_date}")
        util.dump_data(session.participant.participant_id, session.participant.collection_days[0],
                       session.participant.collection_days[-1], sleep_data)
        processed_sleep, sleep_stages = util.process_sleep_response(session.participant.participant_id, sleep_data)
        if processed_sleep is None:
            print(f"No sleep data available for user {user_id}.")
            return None, None
        try:
            sorted = processed_sleep.sort_values(by='date')
            print(sorted)
        except KeyError as e:
            print(f"KeyError: {e} - Ensure 'date' column exists in the DataFrame.")
            return None, None
        return sorted, sleep_stages.sort_values(by=['pid', 'date', 'time'], ascending=[True, True, True])

    def collect_fitbit_data(self, participant_manager: ParticipantManager):
        root_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), '../../'))
        output_path = os.path.join(root_folder, "processed_data")
        os.makedirs(output_path, exist_ok=True)
        active_sessions = participant_manager.get_active_sessions()
        all_participant = []
        all_participant_intraday_steps = []
        all_participant_intraday_hr = []
        all_participant_sleep_stages = []
        for session in active_sessions:
            participant = session.participant
            if participant.participant_id in participant_manager.participants_error:
                log.warning(f"{participant.participant_id} is skipped")
                print(f"{participant.participant_id} is skipped")
                continue
            user_id = session.device.device_id

            print(f"Collecting sleep data for participant {participant.participant_id}")
            try:
                df_sleep, df_sleep_stages = self.collect_sleep_data(session, participant_manager)
                if df_sleep is None:
                    print(f"No sleep data was found for user {user_id}.")
                    continue
                all_participant.append(df_sleep)
                all_participant_sleep_stages.append(df_sleep_stages)
                if "activity" in self.SCOPES or "heartrate" in self.SCOPES:
                    df_steps, df_hr = self.collect_intraday_data(session, participant_manager)
                    if df_steps is not None:
                        all_participant_intraday_steps.append(df_steps)
                    if df_hr is not None:
                        all_participant_intraday_hr.append(df_hr)
            except Exception as ex:
                log.error(f"Error collecting data for participant {participant.participant_id}: {str(ex)}")
                print(f"Error collecting data for participant {participant.participant_id}: {str(ex)}")
                continue
        if not all_participant:
            print(f"No data to process. All participants were skipped/failed")
        else:
            #sleep data and stages
            pd_for_all = pd.concat(all_participant, ignore_index=True)
            pd_ss_all = pd.concat(all_participant_sleep_stages, ignore_index=True)
            pd_for_all.to_csv(os.path.join(output_path, "all_sleep_data.csv"), index=False)
            pd_ss_all.to_csv(os.path.join(output_path," all_fitbit_ss.csv"), index=False)
            #intraday data
            if len(all_participant_intraday_steps) > 0:
                pd_steps_all = pd.concat(all_participant_intraday_steps, ignore_index=True)
                pd_steps_all.to_csv(os.path.join(output_path, "all_step_counts.csv"), index=False)
            if len(all_participant_intraday_hr) > 0:
                pd_hr_all = pd.concat(all_participant_intraday_hr, ignore_index=True)
                pd_hr_all.to_csv(os.path.join(output_path,"all_heart_rate.csv"), index=False)



import fitbit_data_pipeline.Utility as util
import os
import sys

from flask import Flask, request

app = Flask(__name__)

# Flask route to handle the redirect URI after OAuth authorization
project_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))
auth_path = os.path.join(project_root, 'auth_code.txt')
log = util.get_logger()

@app.route('/hello/', methods=['GET'])
def callback():
    # Extract the authorization code from the URL

    code = request.args.get('code')
    state = request.args.get('state')
    log.info(f"Received code: {code}, state: {state}")
    #auth_path = os.path.abspath("../../auth_code.txt")
    print(auth_path)
    log.info(auth_path)
    with open(auth_path, "w") as file:
        log.info("Got here...")
        file.write(f"https://localhost:105/hello/?code={code}&state={state}")
        log.info(f"Auth code written to {auth_path}")
    return f'Authorization successfully received! You may close this tab', 200
    #return request, 200

if __name__ == '__main__':
    app.run(ssl_context='adhoc', port=105)


import logging
import os
import json
import platform
import subprocess

import pandas as pd
from datetime import datetime, timedelta

root_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), '../'))


def get_time(date_time_str):
    """Convert a timestamp string to a time object."""
    return datetime.strptime(date_time_str, '%Y-%m-%dT%H:%M:%S.%f').time()


def dump_data(uid, start_date, end_date, sleep_data):
    """Save sleep data as a JSON file."""
    folder_path = "./raw_sleep_data"
    os.makedirs(folder_path, exist_ok=True)

    file_path = os.path.join(folder_path, f"{uid}_{start_date}_to_{end_date}.json")

    with open(file_path, "w") as json_file:
        json.dump(sleep_data, json_file, indent=4)

    log = get_logger()
    log.info(f"Sleep data saved to {file_path}")
    print(f"Sleep data saved to {file_path}")


def time_diff(time1, time2):
    """Calculate the difference in minutes between two time objects."""
    if isinstance(time1, str):
        time1 = datetime.strptime(time1, "%H:%M").time()
    if isinstance(time2, str):
        time2 = datetime.strptime(time2, "%H:%M").time()

    dt1, dt2 = datetime.combine(datetime.today(), time1), datetime.combine(datetime.today(), time2)

    # Handle cases where time2 is past midnight
    if dt2 < dt1:
        dt2 += timedelta(days=1)

    return (dt2 - dt1).seconds / 60


def process_sleep_response(pid, sleep_data):
    """Process sleep data and return structured DataFrames."""
    all_sleep = []
    all_sleep_stages = []

    for night in sleep_data:
        levels_data = night['levels']['data']


        bed_time, out_of_bed_time = get_time(night['startTime']), get_time(night['endTime'])
        minutesAwake = night['minutesAwake']
        minutesAsleep = night['minutesAsleep']
        timeInBed = night['timeInBed']
        efficiency = night['efficiency']
        type = night['type']
        short_data = night['levels'].get('shortData', [])

        ############## This block contained derived values ########################################
        sleep_start = get_time(levels_data[0]['dateTime']) if(levels_data[0]['level'] != "wake") else get_time(levels_data[1]['dateTime'])
        wake_start = get_time(night['endTime']) if (levels_data[-1]['level'] != 'wake') else get_time(levels_data[-1]['dateTime'])
        SOL = 0 if levels_data[0]['level'] != 'wake' else levels_data[0]['seconds'] / 60
        WASO = night['minutesAwake'] - SOL
        TST = time_diff(sleep_start, wake_start) - WASO
        SE = round((TST / timeInBed),2) * 100 if minutesAsleep else 0
        ############## This block contained derived values ########################################
        # Extract sleep stage summary data
        summary = night['levels'].get('summary', {})
        stage_counts = {stage: summary.get(stage, {}).get('count', 0) for stage in ['wake', 'rem', 'light', 'deep']}
        stage_minutes = {stage: summary.get(stage, {}).get('minutes', 0) for stage in ['wake', 'rem', 'light', 'deep']}

        # Store sleep data
        sleep_df = pd.DataFrame([{
            'pid': pid,
            'date': night['dateOfSleep'],
            'StartTime': bed_time,
            'EndTime': out_of_bed_time,
            'TimeInBed': timeInBed,
            'MinutesAwake': minutesAwake,
            'TotalSleepTime': minutesAsleep,
            **stage_counts,
            **stage_minutes,
            'SleepEfficiency': efficiency,
            'sleep_time': sleep_start,
            'wake_time': wake_start,
            'SOL': SOL,
            'WASO': WASO,
            'TST': TST,
            'SE': SE,
            'LogType': type,
            'isMainSleep': night['isMainSleep'],
        }])

        all_sleep.append(sleep_df)

        if night['type'] == 'stages':
            sleep_stages_df = get_fitbit_epoch(pid, levels_data, short_data)
            all_sleep_stages.append(sleep_stages_df)

    sleep_df = pd.concat(all_sleep, ignore_index=True) if all_sleep else None
    sleep_stages_df = pd.concat(all_sleep_stages, ignore_index=True) if all_sleep_stages else None

    return sleep_df, sleep_stages_df


def get_fitbit_epoch(pid, data, short_data):
    """Convert Fitbit sleep data into 30-second epochs."""
    stage_map = {"wake": "W", "light": "L", "deep": "D", "rem": "R"}
    short_wakes = {datetime.fromisoformat(sh['dateTime']).strftime("%H:%M:%S") for sh in short_data for _ in
                   range(sh['seconds'] // 30)}

    epochs = []
    for entry in data:
        start_time = datetime.fromisoformat(entry['dateTime'])
        stage = stage_map.get(entry['level'], "")

        for _ in range(entry['seconds'] // 30):
            time_str = start_time.strftime("%H:%M:%S")
            if time_str in short_wakes:
                stage = "W"
            epochs.append({"pid": pid, "date": start_time.strftime("%Y-%m-%d"), "time": time_str, "sleep_stage": stage})
            start_time += timedelta(seconds=30)

    return pd.DataFrame(epochs)


def browser_incognito(url):
    if platform.system() == "Windows":
        subprocess.Popen(f'start chrome --new-window --incognito "{url}"', shell=True)


def get_logger(log_file= 'app.log'):
    logger = logging.getLogger(log_file)

    if not logger.handlers:
        logger.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        log_path = os.path.join(root_folder, 'logs')
        file_path = os.path.join(log_path, log_file)
        file_handler = logging.FileHandler(file_path)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    return logger



import csv
import time

from fitbit_data_pipeline.Classes.DataCollector import DataCollector
from fitbit_data_pipeline.Classes.Participant import Participant, Device
from fitbit_data_pipeline.Classes.PManager import ParticipantManager
import config as cfg
import os
import fitbit_data_pipeline.Utility as util

client_id = os.getenv("client_id")
client_secret = os.getenv("client_secret")
redirect_uri = cfg.REDIRECT_URI


def execute():
    root_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), '../'))
    print(f"Root folder: {root_folder}")
    output_path = os.path.join(root_folder, "logs")
    os.makedirs(output_path, exist_ok=True)
    data_collector = DataCollector(client_id, client_secret, redirect_uri)
    participant_manager = ParticipantManager()
    log = util.get_logger()

    try:
        with open(root_folder+'/participants.csv', "r") as file:
            p_data = csv.DictReader(file)
            ##print("I got here")
            log.info("Welcome Here!")
            participants = []
            devices = []
            for p in p_data:
                #print(p["pid"] + p["age"] )
               # print(p["study_period"].split(',') )
                try:
                    participant = Participant(p["pid"], p["age"], p["study_period"].split(','), p["collection_dates"].split(';'))
                    participants.append(participant)
                    device = Device(p["fitbit_id"], p["device_model"])
                    devices.append(device)
                except Exception as e:
                    log.error(f"Skipping - {e}")
            for participant, device in zip(participants, devices):
                try:
                    participant_manager.add_participant(participant)
                    participant_manager.add_device(device)

                    participant_manager.assign_device_to_participant(participant.participant_id,device.device_id)
                    participant.assigned_fitbit = device.device_id
                    data_collector.get_fitbit_oauth(participant, participant_manager)
                    time.sleep(3)
                except Exception as e:
                    participant_manager.participants_error.append(participant.participant_id)
                    log.info(f"Participant {participant.participant_id} returned with error {e} and thus skipping")
            # Collect Data for All Participants
            data_collector.collect_fitbit_data(participant_manager)


            for participant in participants:
                session = participant_manager.get_participant_session(participant.participant_id)
                if session:
                    participant_manager.end_session(session, participant.study_period[1])
    except FileNotFoundError:
        print("Opps! No participant file found. Are you sure you have participants.csv in fitbit_data_pipeline directory?.")

